#include "stdafx.h"
#include "resource.h"
#include "eq_supereq.h"
#include "eq_paramlist.h"

enum {IDC_EQ_EMBEDDED = 666};

enum {N_BANDS=18};


// ===== FORKING NOTE ===== 
// When forking, replace these guids with your own.
// Failure to do so will cause GUID clashes between your component and foobar2000 distribution contents.
static const GUID GUID_dsp_equalizer = { 0x82aef845, 0xdcc3, 0x4da5, { 0x9d, 0x80, 0xe9, 0xa9, 0x72, 0xb2, 0x14, 0xd } };
static const GUID GUID_ui_element_equalizer = { 0x18412765, 0x4342, 0x41bd, { 0x94, 0x16, 0x6d, 0xca, 0x4f, 0x99, 0x93, 0x33 } };
static const GUID GUID_menu_command_equalizer = { 0xc7bbf941, 0x5877, 0x4e11, { 0xa9, 0xbd, 0xbc, 0x19, 0xe6, 0xa6, 0x70, 0x27 } };


struct t_eq_config {
	char bands[N_BANDS];
};

bool operator==(const t_eq_config & cfg1, const t_eq_config & cfg2) {return memcmp(&cfg1,&cfg2,sizeof(t_eq_config)) == 0;}
bool operator!=(const t_eq_config & cfg1, const t_eq_config & cfg2) {return !(cfg1==cfg2);}

static const t_eq_config g_eq_config_default = {20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20};
static const char * const bandLabels[] = {"55 Hz","77 Hz","110 Hz","156 Hz","220 Hz","311 Hz","440 Hz","622 Hz","880 Hz","1.2 kHz","1.8 kHz","2.5 kHz","3.5 kHz","5 kHz","7 kHz","10 kHz","14 kHz","20 kHz"};

static void setup_bands(const t_eq_config & src,double dst[N_BANDS])
{
	int n;
	for(n=0;n<N_BANDS;n++)
	{
		dst[n] = (double)pow(10.0,(src.bands[n]-20)/-20.0);
	}
}

static void g_preset_to_config(t_eq_config & p_config,const dsp_preset & p_preset)
{
	if (p_preset.get_data_size() != sizeof(p_config)) p_config = g_eq_config_default;
	else p_config = * reinterpret_cast<const t_eq_config*>(p_preset.get_data());
}


static void g_config_to_preset(dsp_preset & p_preset,const t_eq_config & p_config)
{
	p_preset.set_owner(GUID_dsp_equalizer);
	p_preset.set_data(&p_config,sizeof(p_config));
}

class dsp_supereq : public dsp_impl_base
{
private:
	pfc::ptr_list_t<supereq_base> eqs;
	unsigned m_last_channels,m_last_channel_config,m_last_sample_rate;
	paramlist paramroot;
	t_eq_config my_eq;
	t_eq_config m_config;

	pfc::array_t<audio_sample> interlace_buf;

	void eq_write_chunk(const audio_chunk * chunk)
	{
		if (!chunk->is_empty()) {
			t_size nch = chunk->get_channels(), samples = chunk->get_sample_count();

			const audio_sample * data = chunk->get_data();

			interlace_buf.grow_size(samples);//deinterlace
			
			
			for(t_size n=0;n<nch;n++) {
				t_size s;
				for(s=0;s<samples;s++) {
					interlace_buf[s] = data[s*nch+n];
				}
				
				eqs[n]->write_samples(interlace_buf.get_ptr(),samples);
			}
		}
	}

	void eq_read_chunk(audio_chunk * chunk)
	{
		t_size total_samples = 0;
		t_size n;

		for(n=0;n<m_last_channels;n++)
		{
			int samples_out = 0;
			audio_sample *data_out = eqs[n]->get_output(&samples_out);
			if (total_samples<(unsigned)samples_out) total_samples=samples_out;
			interlace_buf.grow_size(total_samples*m_last_channels);
			t_size s;
			for(s=0;s<(unsigned)samples_out;s++)
				interlace_buf[s*m_last_channels+n]=data_out[s];
		}

		if (total_samples>0)
		{
			if (chunk==NULL) chunk = insert_chunk(total_samples * m_last_channels);
			if (chunk!=NULL) chunk->set_data(interlace_buf.get_ptr(),total_samples,m_last_channels,m_last_sample_rate,m_last_channel_config);
		}
		else if (chunk!=NULL)
		{
			chunk->reset();
		}
	}
	
public:
	dsp_supereq(const dsp_preset & p_preset) : m_config(g_eq_config_default) {
		g_preset_to_config(m_config,p_preset);
	}

	~dsp_supereq()
	{
		eqs.delete_all();
	}

	static GUID g_get_guid()
	{
		return GUID_dsp_equalizer;
	}

	static void g_get_name(pfc::string_base & p_out) { p_out = "Equalizer";}

	
	bool on_chunk(audio_chunk * p_chunk,abort_callback &)
	{
		if (eqs.get_count()>0 && (m_last_channels!=p_chunk->get_channels() || m_last_channel_config != p_chunk->get_channel_config() || m_last_sample_rate!=p_chunk->get_srate()))
		{
			unsigned n;
			for(n=0;n<m_last_channels;n++)
				eqs[n]->write_samples(0,0);

			eq_read_chunk(0);

			eqs.delete_all();
		}

		m_last_sample_rate = p_chunk->get_srate();
		m_last_channels= p_chunk->get_channels();
		m_last_channel_config = p_chunk->get_channel_config();
		
		if (eqs.get_count()==0)
		{
			unsigned n;
			for(n=0;n<m_last_channels;n++)
				eqs.add_item(new supereq<float>);
			double bands[N_BANDS];
			my_eq = m_config;
			setup_bands(my_eq,bands);
			for(n=0;n<m_last_channels;n++)
				eqs[n]->equ_makeTable(bands,&paramroot,(double)m_last_sample_rate);
		}
		else
		{
			if (m_config != my_eq) {
				my_eq = m_config;
				double bands[N_BANDS];
				setup_bands(my_eq,bands);
				for(unsigned n=0;n<m_last_channels;n++)
					eqs[n]->equ_makeTable(bands,&paramroot,(double)m_last_sample_rate);
			}
		}

		if (eqs.get_count()>0)
		{
			eq_write_chunk(p_chunk);
			eq_read_chunk(p_chunk);
			return !p_chunk->is_empty();
		}
		else return true;
	}

	void on_endoftrack(abort_callback&) {}

	void on_endofplayback(abort_callback &)
	{
		if (eqs.get_count()>0 && m_last_channels==eqs.get_count())
		{
			unsigned n;
			for(n=0;n<m_last_channels;n++)
				eqs[n]->write_samples(0,0);

			eq_read_chunk(0);
			eqs.delete_all();
		}
	}

	virtual double get_latency()
	{
		return eqs.get_count()>0 ? (double)eqs[0]->samples_buffered() / (double)m_last_sample_rate : 0;
	}

	virtual void flush()
	{
		eqs.delete_all();
	}

	bool need_track_change_mark() {return false;}

	static bool g_have_config_popup() {return true;}

	static void g_show_config_popup(const dsp_preset & p_data,HWND p_parent,dsp_preset_edit_callback & p_callback);

	static bool g_get_default_preset(dsp_preset & p_preset)
	{
		g_config_to_preset(p_preset,g_eq_config_default);
		return true;
	}
};

static dsp_factory_t<dsp_supereq> g_dsp_supereq_factory;

static void SavePreset(t_eq_config cfg, const char * path,abort_callback & abort) {
	pfc::string_formatter content;
	for(t_size walk = 0; walk < tabsize(cfg.bands); ++walk) {
		content << (20 - cfg.bands[walk]) << "\r\n";
	}
	
	file::ptr f;
	filesystem::g_open_write_new(f,path,abort);
	f->write_string_raw(content,abort);
}

static void LoadPreset(t_eq_config & cfg, const char * path, abort_callback & abort) {
	
	pfc::string8 content; 
	{
		bool isUTF8; //not really used
		file::ptr f; filesystem::g_open_read(f,path,abort);
		if (f->get_size(abort) > 64*1024 /*sanity check*/) throw exception_io_data();
		text_file_loader::read(f,abort,content,isUTF8);
	}
	pfc::chain_list_v2_t<pfc::string8> lines;
	pfc::splitStringByLines(lines,content);
	t_size cfgWalk = 0;
	for(pfc::const_iterator<pfc::string8> walk = lines.first(); walk.is_valid() && cfgWalk < tabsize(cfg.bands); ++walk) {
		int val = atoi(*walk);
		if (val < -20 || val > 20) throw exception_io_data();
		cfg.bands[cfgWalk++] = (char) ( 20 - val );
	}
	if (cfgWalk < tabsize(cfg.bands)) throw exception_io_data();
}

static void SavePresetQuery(t_eq_config cfg, HWND parent) {
	pfc::string8 path;
	if (uGetOpenFileName(parent,"Equalizer presets (*.feq)|*.feq",0,"feq",NULL,NULL,path,TRUE)) {
		try {
			SavePreset(cfg,pfc::string_formatter() << "file://" << path, abort_callback_dummy());
		} catch(std::exception const & e) {
			popup_message::g_complain("Saving equalizer preset failed",e);
		}
	}
}
static bool LoadPresetQuery(t_eq_config & cfg, HWND parent) {
	pfc::string8 path;
	if (uGetOpenFileName(parent,"Equalizer presets (*.feq)|*.feq",0,"feq",NULL,NULL,path,FALSE)) {
		try {
			LoadPreset(cfg,pfc::string_formatter() << "file://" << path, abort_callback_dummy());
			return true;
		} catch(std::exception const & e) {
			popup_message::g_complain("Loading equalizer preset failed",e);
		}
	}
	return false;
}


static void eq_format(pfc::string_base & out, int val) {
	int dbval;

	dbval = -(val-20);
	pfc::string_formatter temp; temp << dbval;
	out.reset();
	if (temp.length() < 3) out.add_chars(' ',3 - temp.length());
	out << temp;
}

static void EQEnable(const dsp_preset & data) {
	static_api_ptr_t<dsp_config_manager>()->core_enable_dsp(data);
}

static void EQDisable() {
	static_api_ptr_t<dsp_config_manager>()->core_disable_dsp(GUID_dsp_equalizer);
}

static void EQEnable(t_eq_config config) {
	dsp_preset_impl preset;
	g_config_to_preset(preset,config);
	EQEnable(preset);
}


class _DSPConfigNotify {
public:
	virtual void DSPConfigChange(dsp_chain_config const & cfg) {}
};

typedef pfc::instanceTracker<_DSPConfigNotify> DSPConfigNotify;


class CEQPanel : public CWindowImpl<CEQPanel> {
public:
	CEQPanel();
	void CreatePanel(HWND parent,bool showToggle, UINT id = 0,DWORD style = 0, DWORD styleEx = 0);
	BEGIN_MSG_MAP_EX(CEQPanel)
		MSG_WM_SIZE(OnSize)
		MSG_WM_VSCROLL(OnScroll)
		MSG_WM_HSCROLL(OnScroll)
		MSG_WM_SETFONT(OnSetFont)
		MSG_WM_GETFONT(OnGetFont)
		COMMAND_HANDLER_EX(KIDEnabled,BN_CLICKED,OnEnabledToggle)
		COMMAND_HANDLER_EX(KIDZero,BN_CLICKED,OnZero)
		COMMAND_HANDLER_EX(KIDAutoLevel,BN_CLICKED,OnAutoLevel)
		COMMAND_HANDLER_EX(KIDLoadPreset,BN_CLICKED,OnLoadPreset)
		COMMAND_HANDLER_EX(KIDSavePreset,BN_CLICKED,OnSavePreset)
		MSG_WM_GETMINMAXINFO(OnGetMinMaxInfo)
	END_MSG_MAP()

	DECLARE_WND_CLASS_EX(TEXT("{909CC570-5C44-4a42-8B32-CB06B5E18776}"),0,COLOR_BTNFACE)
	void SetEQEnabled(bool state) {m_buttonEnabled.SetCheck(state ? BST_CHECKED : BST_UNCHECKED);}
	bool IsEQEnabled() {return m_buttonEnabled == NULL || m_buttonEnabled.GetCheck() == BST_CHECKED;}
	void SetEQConfig(t_eq_config cfg);
	t_eq_config GrabEQConfig();
	void SetHostModal(bool val) {m_hostIsModal = val;}
protected:
	virtual void OnConfigChanged() {}
private:
	
	
	void OnSize(UINT state, CSize size);
	void OnScroll(UINT nSBCode, UINT nPos, CScrollBar pScrollBar);
	void SetupLabel(CStatic & label, const TCHAR * text);
	void OnEnabledToggle(UINT uNotifyCode, int nID, CWindow wndCtl);
	void OnZero(UINT uNotifyCode, int nID, CWindow wndCtl);
	void OnAutoLevel(UINT uNotifyCode, int nID, CWindow wndCtl);
	void OnLoadPreset(UINT uNotifyCode, int nID, CWindow wndCtl);
	void OnSavePreset(UINT uNotifyCode, int nID, CWindow wndCtl);
	void OnSetFont(HFONT,BOOL);
	void OnGetMinMaxInfo(LPMINMAXINFO lpMMI);
	HFONT OnGetFont();
	int CalcMinWidth();

	enum {KIDEnabled = 1, KIDZero, KIDAutoLevel, KIDLoadPreset, KIDSavePreset, KSliderIDBase};
	CTrackBarCtrl m_sliders[N_BANDS];
	CStatic m_sliderLabels[N_BANDS];
	CStatic m_dbLabels[3];
	CButton m_buttonEnabled, m_buttonZero, m_buttonAutoLevel, m_buttonLoadPreset, m_buttonSavePreset;
	void UpdateLayout();
	CFontHandle m_font;
	const SIZE m_dpi;
	bool m_hostIsModal;
};
void CEQPanel::OnGetMinMaxInfo(LPMINMAXINFO lpMMI) {
	lpMMI->ptMinTrackSize.x = CalcMinWidth();
	lpMMI->ptMinTrackSize.y = MulDiv(200,m_dpi.cy,96);
}
void CEQPanel::OnLoadPreset(UINT uNotifyCode, int nID, CWindow wndCtl) {
	if (!m_hostIsModal && !modal_dialog_scope::can_create()) {
		modal_dialog_scope::poke_existing(); return;
	}
	t_eq_config cfg;
	if (LoadPresetQuery(cfg,*this)) {
		SetEQConfig(cfg); 
		if (IsEQEnabled()) OnConfigChanged();
	}
}

void CEQPanel::OnSavePreset(UINT uNotifyCode, int nID, CWindow wndCtl) {
	if (!m_hostIsModal && !modal_dialog_scope::can_create()) {
		modal_dialog_scope::poke_existing(); return;
	}
	SavePresetQuery(GrabEQConfig(),*this);
}

static BOOL CALLBACK SetFontProc(HWND wnd, LPARAM lp) {
	CWindow(wnd).SetFont((HFONT)lp);
	return TRUE;
}

void CEQPanel::OnSetFont(HFONT font,BOOL) {
	m_font = font;
	::EnumChildWindows(*this,SetFontProc,(LPARAM)font);
	UpdateLayout();
}
HFONT CEQPanel::OnGetFont() {
	return m_font;
}

void CEQPanel::OnZero(UINT uNotifyCode, int nID, CWindow wndCtl) {
	SetEQConfig(g_eq_config_default);
	if (IsEQEnabled()) OnConfigChanged();
}
void CEQPanel::OnAutoLevel(UINT uNotifyCode, int nID, CWindow wndCtl) {
	t_eq_config cfg = GrabEQConfig();
	int peak = cfg.bands[0];
	for(t_size walk = 1; walk < tabsize(cfg.bands); ++walk) {
		if (cfg.bands[walk] < peak) peak = cfg.bands[walk];
	}
	for(t_size walk = 0; walk < tabsize(cfg.bands); ++walk) {
		int val = 20 + cfg.bands[walk] - peak;
		if (val > 40) val = 40;
		cfg.bands[walk] = (char)val;
	}
	SetEQConfig(cfg);
	if (IsEQEnabled()) OnConfigChanged();
}

void CEQPanel::OnEnabledToggle(UINT uNotifyCode, int nID, CWindow wndCtl) {
	bool newState = m_buttonEnabled.GetCheck() == BST_UNCHECKED;
	m_buttonEnabled.SetCheck(newState ? BST_CHECKED : BST_UNCHECKED);
	OnConfigChanged();
}


void CEQPanel::OnScroll(UINT nSBCode, UINT nPos, CScrollBar pScrollBar) {
	if (IsEQEnabled()) OnConfigChanged();
}

void CEQPanel::OnSize(UINT state, CSize size) {
	UpdateLayout();
}

static void DeferHideWindow(HWND wnd, HDWP dwp) {
	if (wnd != NULL) CWindow(wnd).DeferWindowPos(dwp, NULL, 0,0,0,0, SWP_NOZORDER|SWP_NOACTIVATE|SWP_HIDEWINDOW|SWP_NOMOVE|SWP_NOSIZE);
}

static void CheckBoxLayoutHelper(CButton button, HDWP dwp, CSize spacingSize, CRect & rc) {
	if (button.m_hWnd != NULL) {
		CSize ideal;
		{
			pfc::string8 label; uGetWindowText(button,label);
			CWindowDC dc(button);
			SelectObjectScope fontScope( dc, button.GetFont() );
			uGetTextExtentPoint32(dc,label,~0,&ideal);
			ideal.cx += spacingSize.cx * 3;
		}
		rc.left = rc.right + spacingSize.cx; rc.right = rc.left + ideal.cx;
		if (dwp != NULL) {
			CRect myRC(rc);
			button.DeferWindowPos(dwp, NULL, myRC.left, myRC.top, myRC.Width(), myRC.Height(), SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
		}
	}
}

static void ButtonLayoutHelper(CButton button, HDWP dwp, CSize spacingSize, CRect & rc) {
	if (button.m_hWnd != NULL) {
		CSize ideal; WIN32_OP_D( button.GetIdealSize(&ideal) );
		rc.left = rc.right + spacingSize.cx; rc.right = rc.left + ideal.cx;
		if (dwp != NULL) {
			CRect myRC(rc);
			if (ideal.cy < myRC.Height()) myRC.top = myRC.CenterPoint().y - ideal.cy / 2; myRC.bottom = myRC.top + ideal.cy;
			button.DeferWindowPos(dwp, NULL, myRC.left, myRC.top, myRC.Width(), myRC.Height(), SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
		}
	}
}

int CEQPanel::CalcMinWidth() {
	CSize spacingSize;
	{
		CWindowDC dc(*this);
		SelectObjectScope fontScope(dc,m_font);
		uGetTextExtentPoint32(dc,"#",~0,&spacingSize);
	}
	{
		const HDWP dwp = NULL;
		CRect rc;
		rc.top = rc.bottom = 0;
		rc.left = rc.right = 0;
		CheckBoxLayoutHelper(m_buttonEnabled,dwp,spacingSize,rc);
		ButtonLayoutHelper(m_buttonZero,dwp,spacingSize,rc);
		ButtonLayoutHelper(m_buttonAutoLevel,dwp,spacingSize,rc);
		ButtonLayoutHelper(m_buttonLoadPreset,dwp,spacingSize,rc);
		ButtonLayoutHelper(m_buttonSavePreset,dwp,spacingSize,rc);
		return rc.right + spacingSize.cx;
	}
}

void CEQPanel::UpdateLayout() {
	CRect rcClient;
	WIN32_OP_D(GetClientRect(&rcClient));
	const int clWidth = rcClient.Width(); const int clHeight = rcClient.Height();
	const int labelHeight = uGetFontHeight( m_font );

	CSize dbLabelSize, spacingSize;
	{
		CWindowDC dc(*this);
		SelectObjectScope fontScope(dc,m_font);
		uGetTextExtentPoint32(dc,"+20dB",~0,&dbLabelSize);
		uGetTextExtentPoint32(dc,"#",~0,&spacingSize);
	}
	const int rMargin = dbLabelSize.cx;
	const int bMargin = labelHeight * 4;
	
	HDWP dwp = BeginDeferWindowPos(N_BANDS * 2 + 8);
	if (rMargin < clWidth && bMargin < clHeight) {
		const int tbAreaWidth = clWidth - rMargin;
		const int tbAreaHeight = clHeight - bMargin;
		const int tbWidth = tbAreaWidth / N_BANDS;
		for(t_size walk = 0; walk < N_BANDS; ++walk) {
			const int xPos = tbAreaWidth * walk / N_BANDS;
			if (m_sliders[walk] != NULL) {
				CRect rc; rc.top = 0; rc.bottom = tbAreaHeight;
				rc.left = xPos; rc.right = rc.left + tbWidth;
				m_sliders[walk].DeferWindowPos(dwp, NULL, rc.left, rc.top, rc.Width(), rc.Height(), SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
			}
			if (m_sliderLabels[walk] != NULL) {
				CRect rc; rc.top = tbAreaHeight + (walk % 2) * labelHeight; rc.bottom = rc.top + labelHeight;
				rc.left = xPos; rc.right = rc.left + tbWidth * 2;
				m_sliderLabels[walk].DeferWindowPos(dwp, NULL, rc.left, rc.top, rc.Width(), rc.Height(), SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
			}
		}
		
		{
			CRect rc;
			rc.top = 0; rc.bottom = rc.top + labelHeight;
			rc.left = tbAreaWidth; rc.right = rc.left + rMargin;
			m_dbLabels[0].DeferWindowPos(dwp, NULL, rc.left, rc.top, rc.Width(), rc.Height(), SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
			rc.top = tbAreaHeight / 2 - labelHeight / 2; rc.bottom = rc.top + labelHeight;
			m_dbLabels[1].DeferWindowPos(dwp, NULL, rc.left, rc.top, rc.Width(), rc.Height(), SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
			rc.top = tbAreaHeight - labelHeight; rc.bottom = rc.top + labelHeight;
			m_dbLabels[2].DeferWindowPos(dwp, NULL, rc.left, rc.top, rc.Width(), rc.Height(), SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
		}

		{
			CRect rc;
			rc.top = tbAreaHeight + 2 * labelHeight; rc.bottom = clHeight;
			rc.left = rc.right = 0;
			CheckBoxLayoutHelper(m_buttonEnabled,dwp,spacingSize,rc);
			ButtonLayoutHelper(m_buttonZero,dwp,spacingSize,rc);
			ButtonLayoutHelper(m_buttonAutoLevel,dwp,spacingSize,rc);
			ButtonLayoutHelper(m_buttonLoadPreset,dwp,spacingSize,rc);
			ButtonLayoutHelper(m_buttonSavePreset,dwp,spacingSize,rc);
		}
	} else {
		for(t_size walk = 0; walk < N_BANDS; ++walk) {
			DeferHideWindow(m_sliders[walk],dwp);
			DeferHideWindow(m_sliderLabels[walk],dwp);
		}
		for(t_size walk = 0; walk < 3; ++walk) {
			DeferHideWindow(m_dbLabels[walk],dwp);
		}
		DeferHideWindow(m_buttonEnabled,dwp);

	}
	EndDeferWindowPos(dwp);
}

static const DWORD DefCtrlStyle = CControlWinTraits::GetWndStyle(0) | WS_TABSTOP;
static const DWORD DefLabelStyle = CControlWinTraits::GetWndStyle(0);

void CEQPanel::SetupLabel(CStatic & label, const TCHAR * text) {
	WIN32_OP_D( label.Create(*this,NULL, text, DefLabelStyle | SS_LEFT | SS_ENDELLIPSIS) != NULL );
}

CEQPanel::CEQPanel() 
: m_dpi(QueryScreenDPIEx()), m_hostIsModal()
{
}

void CEQPanel::CreatePanel(HWND parent, bool showToggle, UINT id,DWORD style, DWORD styleEx) {
	WIN32_OP( Create(parent, NULL, NULL, style, styleEx | WS_EX_CONTROLPARENT, id) != NULL );	
	
	for(t_size walk = 0; walk < N_BANDS; ++walk) {
		WIN32_OP_D( m_sliders[walk].Create(*this,NULL, NULL, DefCtrlStyle | TBS_VERT | TBS_BOTH | TBS_NOTICKS, 0, (UINT)(KSliderIDBase + walk)) != NULL );
		m_sliders[walk].SetRange(0,40,FALSE);
		m_sliders[walk].SetPageSize(1);
		m_sliders[walk].SetPos(20);

		SetupLabel(m_sliderLabels[walk], pfc::stringcvt::string_os_from_utf8(bandLabels[walk]));
	}
	SetupLabel(m_dbLabels[0], TEXT("+20dB"));
	SetupLabel(m_dbLabels[1], TEXT("+0dB"));
	SetupLabel(m_dbLabels[2], TEXT("-20dB"));

	if (showToggle) {
		WIN32_OP_D( m_buttonEnabled.Create(*this, NULL, TEXT("Equalizer enabled"), DefCtrlStyle | BS_CHECKBOX | BS_TEXT | BS_VCENTER, 0, KIDEnabled) != NULL );
	}

	WIN32_OP_D( m_buttonZero.Create(*this, NULL, TEXT("Zero all"), DefCtrlStyle | BS_TEXT | BS_VCENTER, 0, KIDZero) != NULL );
	WIN32_OP_D( m_buttonAutoLevel.Create(*this, NULL, TEXT("Auto level"), DefCtrlStyle | BS_TEXT | BS_VCENTER, 0, KIDAutoLevel) != NULL );
	WIN32_OP_D( m_buttonLoadPreset.Create(*this, NULL, TEXT("Load preset"), DefCtrlStyle | BS_TEXT | BS_VCENTER, 0, KIDLoadPreset) != NULL );
	WIN32_OP_D( m_buttonSavePreset.Create(*this, NULL, TEXT("Save preset"), DefCtrlStyle | BS_TEXT | BS_VCENTER, 0, KIDSavePreset) != NULL );
}

t_eq_config CEQPanel::GrabEQConfig() {
	t_eq_config ret;
	for(t_size walk = 0; walk < N_BANDS; ++walk) {
		if (m_sliders[walk] != NULL) {
			ret.bands[walk] = (char)m_sliders[walk].GetPos();
		} else {
			ret.bands[walk] = 20;
		}
	}
	return ret;
}
void CEQPanel::SetEQConfig(t_eq_config cfg) {
	for(t_size walk = 0; walk < N_BANDS; ++walk) {
		if (m_sliders[walk] != NULL) {
			m_sliders[walk].SetPos(cfg.bands[walk]);
		}
	}
}

// UI element
class ui_element_instance_eq : public CEQPanel, public ui_element_instance, private DSPConfigNotify {
public:
	ui_element_instance_eq(ui_element_config::ptr cfg, ui_element_instance_callback::ptr callback);
	void initialize_window(HWND parent);
	void set_configuration(ui_element_config::ptr data) {}
	ui_element_config::ptr get_configuration() {return g_get_default_configuration();}
	static ui_element_config::ptr g_get_default_configuration() {return ui_element_config::g_create_empty(g_get_guid());}
	static GUID g_get_guid() {
		return GUID_ui_element_equalizer;
	}

	static GUID g_get_subclass() {return ui_element_subclass_utility;}
	static void g_get_name(pfc::string_base & out) { out = "Equalizer";}
	static const char * g_get_description() { return "Controls equalizer DSP.";}

	void notify(const GUID & p_what, t_size p_param1, const void * p_param2, t_size p_param2size);

	BEGIN_MSG_MAP_EX(ui_element_instance_eq)
		CHAIN_MSG_MAP(CEQPanel)
	END_MSG_MAP()
private:
	void ApplyCoreSettings();
	void DSPConfigChange(dsp_chain_config const & cfg);
	void OnConfigChanged();

	bool m_ownUpdate;
protected:
	const ui_element_instance_callback::ptr m_callback;
};

void ui_element_instance_eq::ApplyCoreSettings() {
	dsp_preset_impl preset;
	if (static_api_ptr_t<dsp_config_manager>()->core_query_dsp(GUID_dsp_equalizer,preset)) {
		SetEQEnabled(true);
		t_eq_config cfg;
		g_preset_to_config(cfg,preset);
		SetEQConfig(cfg);
	} else {
		SetEQEnabled(false);
	}
}


void ui_element_instance_eq::DSPConfigChange(dsp_chain_config const & cfg) {
	if (!m_ownUpdate && m_hWnd != NULL) {
		ApplyCoreSettings();
	}
}

void ui_element_instance_eq::initialize_window(HWND parent) {
	CreatePanel(parent,true,0,0,WS_EX_STATICEDGE);
	SetFont(m_callback->query_font());
	ApplyCoreSettings();
}
ui_element_instance_eq::ui_element_instance_eq(ui_element_config::ptr cfg, ui_element_instance_callback::ptr callback) 
: m_callback(callback), m_ownUpdate()
{
}

void ui_element_instance_eq::notify(const GUID & p_what, t_size p_param1, const void * p_param2, t_size p_param2size) {
	if (p_what == ui_element_notify_font_changed) {
		SetFont(m_callback->query_font());
		m_callback->on_min_max_info_change();
	}
}

void ui_element_instance_eq::OnConfigChanged() {
	pfc::vartoggle_t<bool> ownUpdate(m_ownUpdate,true);
	if (IsEQEnabled()) {
		::EQEnable(GrabEQConfig());
	} else {
		::EQDisable();
	}
}

class ui_element_eq : public ui_element_impl_withpopup<ui_element_instance_eq> {
public:
	GUID get_menu_command_id() {
		return GUID_menu_command_equalizer;
	}
};
static service_factory_single_t<ui_element_eq> g_ui_element_eq_factory;

class dsp_config_callback_dispatch : public dsp_config_callback {
public:
	void on_core_settings_change(const dsp_chain_config & p_newdata) {
		for(pfc::const_iterator<DSPConfigNotify*> walk = DSPConfigNotify::instanceList().first(); walk.is_valid(); ++walk) {
			(*walk)->DSPConfigChange(p_newdata);
		}
	}
};

static service_factory_single_t<dsp_config_callback_dispatch> g_dsp_config_callback_dispatch_factory;

class CEQConfigPopupContent : public CEQPanel {
public:
	CEQConfigPopupContent(dsp_preset_edit_callback & callback) : m_callback(callback) {}

private:
	void OnConfigChanged() {
		dsp_preset_impl temp; g_config_to_preset(temp,this->GrabEQConfig());
		m_callback.on_preset_changed(temp);
	}
	dsp_preset_edit_callback & m_callback;
};

class CEQConfigPopup : public CDialogImpl<CEQConfigPopup> {
public:
	enum {IDD = IDD_CONFIG_EQ_POPUP};
	CEQConfigPopup(dsp_preset_edit_callback & callback, const dsp_preset & initVal) : m_initVal(initVal), m_callback(callback), m_panel(callback) {}
	BEGIN_MSG_MAP_EX(CEQConfigPopup)
		MSG_WM_INITDIALOG(OnInitDialog)
		MSG_WM_SIZE(OnSize)
		COMMAND_HANDLER_EX(IDCANCEL,BN_CLICKED,OnOKCancel)
		COMMAND_HANDLER_EX(IDOK,BN_CLICKED,OnOKCancel)
		MSG_WM_GETMINMAXINFO(OnGetMinMaxInfo)
	END_MSG_MAP()
private:
	LRESULT OnInitDialog(HWND, LPARAM);
	void ApplyLayout();
	LRESULT OnSize(UINT,CSize);
	void OnOKCancel(UINT uNotifyCode, int nID, CWindow wndCtl) {EndDialog(nID);}
	void OnGetMinMaxInfo(LPMINMAXINFO lpMMI);

	CEQConfigPopupContent m_panel;
	const dsp_preset_impl m_initVal;
	dsp_preset_edit_callback & m_callback;
	CSize m_spacing;
	CButton m_buttonOK, m_buttonCancel;
	modal_dialog_scope m_modalScope;
};

void CEQConfigPopup::OnGetMinMaxInfo(LPMINMAXINFO lpMMI) {
	MINMAXINFO mi = {};
	m_panel.SendMessage(WM_GETMINMAXINFO, 0, (LPARAM) &mi);
	CRect rcCancel;
	WIN32_OP_D( m_buttonCancel.GetWindowRect(rcCancel) ); 
	CRect rc(0, 0, mi.ptMinTrackSize.x + m_spacing.cx * 2, mi.ptMinTrackSize.y + m_spacing.cy * 2 + rcCancel.Height());
	AdjustWindowRectEx(rc, GetStyle(), FALSE, GetExStyle());
	lpMMI->ptMinTrackSize.x = rc.Width();
	lpMMI->ptMinTrackSize.y = rc.Height();

	
}

LRESULT CEQConfigPopup::OnSize(UINT,CSize) {
	ApplyLayout(); return 0;
}

void CEQConfigPopup::ApplyLayout() {
	CRect ok, cancel, client, panel;
	m_buttonOK.GetWindowRect(&ok); m_buttonCancel.GetWindowRect(&cancel);
	GetClientRect(&client);
	
	
	cancel.MoveToXY(client.right - cancel.Width() - m_spacing.cx, client.bottom - cancel.Height() - m_spacing.cy);
	ok.MoveToXY(cancel.left - m_spacing.cx - ok.Width(), cancel.top);
	panel.left = m_spacing.cx; panel.right = client.right - m_spacing.cx;
	panel.top = m_spacing.cy;
	panel.bottom = cancel.top - m_spacing.cy;
	HDWP dwp = ::BeginDeferWindowPos(3);
	::DeferWindowPos(dwp,m_panel,NULL,panel.left,panel.top,panel.Width(),panel.Height(),SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
	::DeferWindowPos(dwp,m_buttonOK,NULL,ok.left,ok.top,ok.Width(),ok.Height(),SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
	::DeferWindowPos(dwp,m_buttonCancel,NULL,cancel.left,cancel.top,cancel.Width(),cancel.Height(),SWP_NOZORDER|SWP_NOACTIVATE|SWP_SHOWWINDOW|SWP_NOCOPYBITS);
	::EndDeferWindowPos(dwp);
}

LRESULT CEQConfigPopup::OnInitDialog(HWND, LPARAM) {
	m_modalScope.initialize(*this);
	try {
		m_panel.SetHostModal(true);
		m_panel.CreatePanel(*this,false,IDC_EQ_EMBEDDED);
		m_panel.SetFont((HFONT)GetStockObject(DEFAULT_GUI_FONT));

		{
			t_eq_config cfg; g_preset_to_config(cfg,m_initVal);
			m_panel.SetEQConfig(cfg);
		}

	} catch(std::exception const & e) {
		popup_message::g_complain("Dialog creation failed",e);
		EndDialog(IDCANCEL);
		return 0;
	}
	m_buttonOK = GetDlgItem(IDOK); m_buttonCancel = GetDlgItem(IDCANCEL);

	{
		CRect client, button;
		GetClientRect(&client);
		ClientToScreen(&client);
		m_buttonCancel.GetWindowRect(&button);
		m_spacing.cx = client.right - button.right; m_spacing.cy = client.bottom - button.bottom;
	}

	ApplyLayout();
	return 0;
}

void dsp_supereq::g_show_config_popup(const dsp_preset & p_data,HWND p_parent,dsp_preset_edit_callback & p_callback) {
	core_api::ensure_main_thread();
	CEQConfigPopup dialog(p_callback,p_data);
	if (dialog.DoModal(p_parent) != IDOK) p_callback.on_preset_changed(p_data);
}

// Version information.

// ===== FORKING NOTE ===== 
// Insert your component name, version and credits here.
DECLARE_COMPONENT_VERSION(
"Equalizer",
"1.1",
"SuperEQ library (c) 2000-2002 Naoki Shibata\n"
"foobar2000 equalizer component (c) 2002-2016 Peter Pawlowski\n"
"\n"
"Distributed under the terms of GNU LGPL v2.1\n"
"Source code can be obtained from:\n"
"http://www.foobar2000.org/equalizer"
);



// Component file name check (prevent user side renaming - for debugging and crash report autotagging sanity).

// ===== FORKING NOTE ===== 
// Insert your new DLL name here.
VALIDATE_COMPONENT_FILENAME("foo_dsp_eq.dll");




// SPECIAL FIX:
// This component is for fb2k 1.2 and newer only. Make it not work in older because it duplicates standard DSP array code there.

// ===== FORKING NOTE ===== 
// remove this code when forking
class validator_fb2kver : public component_installation_validator {
public:
	bool is_installed_correctly() {
		return static_api_ptr_t<core_version_info_v2>()->test_version(1,2,0,0);
	}
};

static service_factory_single_t<validator_fb2kver> g_validator_fb2kver_factory;
// End SPECIAL FIX
